/*
 * flash-tools - a set of tools for flashing handsets
 * Copyright (C) 2005 Jeremy Lain <jeremy.laine@m4x.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include "p2k/p2klib.h"
#include "p2k/p2kdev.h"
#include "kbdhit.h"

#define PROGNAME "datalog-tool"

void usage()
{
  fprintf(stderr, "Syntax:\n");
  fprintf(stderr, "  " PROGNAME " [options]\n\n");
  fprintf(stderr, "Mandatory options:\n");
  fprintf(stderr, "  -o <file>     write the captured data to <file>\n");
  fprintf(stderr, "  -s            write the captured data to the standard output\n");
}


int datalog_read(p2k_dev_handle *dev_handle, FILE *fp)
{
  char read_bytes[FLASH_INPUT_BUFFER];
  int read_len;

  if (!(read_len = p2kdev_read(dev_handle, read_bytes, FLASH_INPUT_BUFFER, P2K_TIMEOUT)))
  {
    fprintf(stderr, "Read error, bailing.\n");
    return RET_ERR;  
  }
  fwrite(read_bytes, read_len, 1, fp);

  return RET_OK;
}


int main(const int argc, char* argv[])
{
  struct usb_device *devs[P2K_DEVICE_MAX];
  size_t nbdevs = 0;
  p2k_dev_handle p2k_handle;
  char *outputfile = NULL;
  int c;

  enum { OUT_NONE, OUT_STDOUT, OUT_FILE } output = OUT_NONE;

  fprintf(stderr, "[ This is " PROGNAME " v" VERSION ", a Motorola P2k handset logging tool ]\n\n");

  // check command line options
  while ((c = getopt (argc, argv, "ho:s")) != -1)
  {
  
    switch (c)
    {
      case 'h':
        usage();
        return 0;
        break;
      case 'o':
        output = OUT_FILE;
        outputfile = optarg;
        break;
      case 's':
        output = OUT_STDOUT;
        break;
    }
  }
  if (output == OUT_NONE)
  {
    fprintf(stderr, "No output was specified!\n\n");
    usage();
    return 1;
  }

  usb_init();
  usb_find_busses();
  usb_find_devices();

  // find modem device  
  if ((nbdevs = p2kdev_find(P2K_VENDOR_MOTOROLA, P2K_PRODUCT_MODEM, &devs[0], P2K_DEVICE_MAX)) > 0)
  {
    fprintf(stderr, "-> Switching USB modem to USB device..\n");
    p2kdev_modem2composite(devs[0]);
    sleep(4);
    usb_find_busses();
    usb_find_devices();
    fprintf(stderr, "\n");
  }

  // find logging device  
  if (!(nbdevs = p2kdev_find(P2K_VENDOR_MOTOROLA, P2K_PRODUCT_COMPOSITE, &devs[0], P2K_DEVICE_MAX)))
  {    
    fprintf(stderr, "Could not find any data logging devices, aborting.\n");
    return 1;
  }

  // select the first flash devices
  fprintf(stderr, "-> Found %i data logging device(s), using first one.\n", nbdevs);

  // open flash device
  if (!p2kdev_open_datalog(&p2k_handle, devs[0]))
  {
    fprintf(stderr, "Could not open USB device, aborting.\n");
    return 1;
  }

  // open output file
  FILE *fp = stdout;
  if (outputfile != NULL)
  {
    fprintf(stderr, "\n-> Logging to file '%s'\n", outputfile);  
    if (!(fp = fopen(outputfile, "wb")))
    {
      fprintf(stderr, "Could not open '%s' for writing, aborting.\n", outputfile);
      p2kdev_close(&p2k_handle);  
      return 1;
    }
  } else {
    fprintf(stderr, "\n-> Logging to standard output\n");
  }

  // read data until the user resses a key
  fprintf(stderr, "Press ENTER to stop logging..\n");
  tty_cbreak(STDIN_FILENO);
  while (!kbdhit())
  {
    datalog_read(&p2k_handle, fp);
  }
  tty_reset(STDIN_FILENO);

  // close output file
  if (outputfile != NULL)
  {
    fclose(fp);
  }

  // close flash device
  p2kdev_close(&p2k_handle);  

  return 0;
}
