/*
 * flash-tools - a set of tools for flashing handsets
 * Copyright (C) 2005 Jeremy Lain <jeremy.laine@m4x.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h> 
#include "p2klib.h"
#include "flashcmd.h"


/* LOW-LEVEL FUNCTIONS */

/** Check a response against and expected answer.
 *
 * @param resp_bytes
 * @param check_bytes
 * @param check_len
 */
int flash_checkresp(const char *resp_bytes, const size_t resp_len, const char *check_bytes, const size_t check_len)
{
   size_t i;
   
   if ((resp_len < check_len) || !resp_bytes || !check_bytes)
   {
     return RET_ERR;
   }
     
   for (i = 0; i < check_len; i++)
   {
     if (resp_bytes[i] != check_bytes[i])
     {
       printf("Invalid command reponse : \n");
       hexprint(resp_bytes, resp_len);
       return RET_ERR;
     }
   }
     
   return RET_OK;
}


/** Send a command to a P2k flash device.
 */
int flash_cmd(p2k_dev_handle *dev_handle, char* bytes, size_t size, char *read_bytes, int *read_len, const int timeout)
{
  if (!p2kdev_write(dev_handle, bytes, size, timeout))
  {
    printf("Write error, bailing.\n");
    return RET_ERR;
  }
  
  if (!(*read_len = p2kdev_read(dev_handle, read_bytes, FLASH_INPUT_BUFFER, timeout)))
  {
    printf("Read error, bailing.\n");
    return RET_ERR;  
  }         
 
  return RET_OK;
}


/** Send a simple command to a P2k flash device.
 */
int flash_cmd_simple(p2k_dev_handle *dev_handle, const char *cmd, char *read_bytes, int *read_len, const int timeout)
{
  char mycmd[256];

#ifdef INFO
  printf("-> Sending %s command\n", cmd);
#endif
  
  sprintf(mycmd, "%c%s%c", FCMD_BEG, cmd, FCMD_END);
  return flash_cmd(dev_handle, mycmd, strlen(mycmd), read_bytes, read_len, timeout);
}


/* MIDLEVEL FUNCTIONS */



/** Send the "address" command.
 */
int flash_cmd_ADDR(p2k_dev_handle *dev_handle, const size_t naddr)
{
  char addr[9];    
  char cmd_addr[32];        
  unsigned char cksum = 0;
  size_t i;
  
  char resp_bytes[FLASH_INPUT_BUFFER];
  char check_bytes[FLASH_INPUT_BUFFER];
  int resp_len = 0;

  sprintf(addr, "%.8X", naddr);

  // checksum calculation
  for (i = 0; i <8 ; i++)
  {
    cksum += (unsigned char)addr[i];
  }
  
  sprintf(cmd_addr,"%cADDR%c%.8X%.2X%c", FCMD_BEG, FCMD_SEP, naddr, cksum, FCMD_END);
    
#ifdef DEBUG
  printf("-> Sending ADDR command\n");
#endif    
  if (!flash_cmd(dev_handle, cmd_addr, 17, resp_bytes, &resp_len, P2K_TIMEOUT))
    return RET_ERR;

  // check response
  sprintf(check_bytes, "%cACK", FCMD_BEG);  
  if (!flash_checkresp(resp_bytes, resp_len, check_bytes, strlen(check_bytes)))
    return RET_ERR;

  return RET_OK;
}


/** Send the BIN command.
 */
int flash_cmd_BIN(p2k_dev_handle *dev_handle, char *data, const size_t datalen)
{
  char buffer[FLASH_BLOCK + 9];
  size_t bufferlen;
  size_t j;
  unsigned char cksum = 0;
  
  char resp_bytes[FLASH_INPUT_BUFFER];
  char check_bytes[FLASH_INPUT_BUFFER];  
  int resp_len = 0;
      
  memcpy(buffer + 7, data, datalen);
    
  sprintf(buffer, "%cBIN%c", FCMD_BEG, FCMD_SEP);    
  buffer[5] = (datalen >> 8) & 0xFF;
  buffer[6] = datalen & 0xFF;
    
  bufferlen = datalen + 9;

  // checksum calculation
  for (j = 5; j < 7 + datalen; j++)
  {
    cksum += (unsigned char)buffer[j];
  }
        
  buffer[7 + datalen] = cksum;
  buffer[8 + datalen] = FCMD_END;

//   // send command
#ifdef DEBUG    
    printf("-> Sending BIN command\n");    
#endif  
  if (!flash_cmd(dev_handle, buffer, bufferlen, resp_bytes, &resp_len, P2K_TIMEOUT))
    return RET_ERR;

  // check response
  sprintf(check_bytes, "%cACK", FCMD_BEG);  
  if (!flash_checkresp(resp_bytes, resp_len, check_bytes, strlen(check_bytes)))
    return RET_ERR;
      
  return RET_OK;
}


/** Send ERASE command
 */
int flash_cmd_ERASE(p2k_dev_handle *flash_handle)
{
  char resp_bytes[FLASH_INPUT_BUFFER];
  char check_bytes[FLASH_INPUT_BUFFER];  
  int resp_len = 0;

  char cmd_erase[64];

  sprintf(cmd_erase, "%cERASE%c%c%c", FCMD_BEG, FCMD_SEP, 0x80, FCMD_END);
    
  // send command
#ifdef INFO  
  printf("-> Sending ERASE command\n");
#endif   
  if (!flash_cmd(flash_handle, cmd_erase, 9, resp_bytes, &resp_len, P2K_TIMEOUT_LONG))
    return RET_ERR;

  // check response
  sprintf(check_bytes, "%cACK", FCMD_BEG);  
  if (!flash_checkresp(resp_bytes, resp_len, check_bytes, strlen(check_bytes)))
    return RET_ERR;
        
  return RET_OK;   
}
 
 
/** Send a JUMP command
 */
int flash_cmd_JUMP(p2k_dev_handle *flash_handle, const size_t addr)
{
  char jaddr[9];    
  unsigned char cksum = 0;
  char cmd_jump[64];  
  size_t k;

  char resp_bytes[FLASH_INPUT_BUFFER];
  char check_bytes[FLASH_INPUT_BUFFER];  
  int resp_len = 0;
   
  sprintf(jaddr, "%.8X", addr);

  // checksum calculation
  for (k = 0; k < 8 ; k++)
  {
    cksum += (unsigned char)jaddr[k];
  }
  
  // send command
#ifdef INFO
  printf("-> Sending JUMP command\n");  
#endif    
  sprintf(cmd_jump,"%cJUMP%c%.8X%.2X%c", FCMD_BEG, FCMD_SEP, addr, cksum, FCMD_END);
  if (!flash_cmd(flash_handle, cmd_jump, 17, resp_bytes, &resp_len, P2K_TIMEOUT))
    return RET_ERR;
    
  // check response
  sprintf(check_bytes, "%cACK", FCMD_BEG);  
  if (!flash_checkresp(resp_bytes, resp_len, check_bytes, strlen(check_bytes)))
    return RET_ERR; 
    
  return RET_OK;
}


/** Restart the phone
 */
int flash_cmd_RESTART(p2k_dev_handle *flash_handle)
{  
  char mycmd[16];

  // send command
#ifdef INFO
  printf("-> Sending RESTART command\n");
#endif  
  sprintf(mycmd, "%cRESTART%c", FCMD_BEG, FCMD_END);  
  if (!p2kdev_write(flash_handle, mycmd, strlen(mycmd), P2K_TIMEOUT))
  {
    printf("Write error, bailing.\n");
    return RET_ERR;
  }      
  return RET_OK;  
}


/** Send RQCS command
 */
int flash_cmd_RQCS(p2k_dev_handle *flash_handle)
{
  return flash_cmd_RQxx(flash_handle, "RQCS", "RSCS", P2K_TIMEOUT_LONG);
}  


/** Send RQHW command
 */
int flash_cmd_RQHW(p2k_dev_handle *flash_handle)
{
  return flash_cmd_RQxx(flash_handle, "RQHW", "RSHW", P2K_TIMEOUT);
}  


/** Send RQRC command
 */
int flash_cmd_RQRC(p2k_dev_handle *flash_handle, size_t flashbase, size_t donebytes)
{
  char cmd_store[32];        
  sprintf(cmd_store,"RQRC%c%.8X,%.8X", FCMD_SEP, flashbase, flashbase + donebytes - 1);  
  
  return flash_cmd_RQxx(flash_handle, cmd_store, "RSRC", P2K_TIMEOUT);
}


/** Send RQRCS command
 */
int flash_cmd_RQRCS(p2k_dev_handle *flash_handle, size_t addrbase, size_t bytes)
{
  char cmd_csum[32];
  sprintf(cmd_csum,"RQRCS%c%.8X,%.8X", FCMD_SEP, addrbase, addrbase + bytes - 1);
  
  return flash_cmd_RQxx(flash_handle, cmd_csum, "RSRCS", P2K_TIMEOUT_LONG);
}


/** Send RQVN command
 */
int flash_cmd_RQSF(p2k_dev_handle *flash_handle)
{
  // TODO : find why we are getting ERR 0x1e 0x85 response
  return flash_cmd_RQxx(flash_handle, "RQSF", "", P2K_TIMEOUT);
}  


/** Send RQSN command
 */
int flash_cmd_RQSN(p2k_dev_handle *flash_handle)
{
  return flash_cmd_RQxx(flash_handle, "RQSN", "RSSN", P2K_TIMEOUT);
}  

  
/** Send RQVN command
 */
int flash_cmd_RQVN(p2k_dev_handle *flash_handle)
{
  return flash_cmd_RQxx(flash_handle, "RQVN", "RSVN", P2K_TIMEOUT);
}  


/** Send RQxx command
 */
int flash_cmd_RQxx(p2k_dev_handle *flash_handle, const char *cmd, const char *resp, const int timeout)
{
  char resp_bytes[FLASH_INPUT_BUFFER];
  int resp_len = 0;
  char check_bytes[FLASH_INPUT_BUFFER];  
    
  // send command
  if (!flash_cmd_simple(flash_handle, cmd, resp_bytes, &resp_len, timeout))
    return RET_ERR;

  // check response
  sprintf(check_bytes, "%c%s", FCMD_BEG, resp);    
  if (!flash_checkresp(resp_bytes, resp_len, check_bytes, strlen(check_bytes)))
    return RET_ERR;
  
  return RET_OK;
}  


/** Send UPGRADE command
 */
int flash_cmd_UPGRADE(p2k_dev_handle *flash_handle)
{  
  char resp_bytes[FLASH_INPUT_BUFFER];
  int resp_len = 0;
  char cmd_upgrade[64];

  // send command  
#ifdef INFO   
  printf("-> Sending UPGRADE command\n");    
#endif  
  sprintf(cmd_upgrade, "%cUPGRADE%cREQ_STAT%c%c", FCMD_BEG, FCMD_SEP, 0x00, FCMD_END);
  if (!flash_cmd(flash_handle, cmd_upgrade, 19, resp_bytes, &resp_len, P2K_TIMEOUT))
    return RET_ERR;
  
  // check response
  // TODO: find out why we are getting an ERR response
  /*
  sprintf(check_bytes, "%cACK", FCMD_BEG);  
  if (!flash_checkresp(resp_bytes, resp_len, check_bytes, strlen(check_bytes)))
    return RET_ERR;  
  */
  
  return RET_OK;
}

