/*
 * flash-tools - a set of tools for flashing handsets
 * Copyright (C) 2005 Jeremy Lain <jeremy.laine@m4x.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <sys/stat.h> 
#include "flashops.h"
#include "flashcmd.h"
#include "srecord.h"


/** Calculate checksums.
 */
int flash_checksum(p2k_dev_handle *dev_handle, flashcat_entry* codegroups, size_t cgcnt)
{  
  size_t i;
  
  if (!flash_cmd_RQCS(dev_handle))
    return RET_ERR;
    
  for (i = 0; i < cgcnt; i++)
  {
    struct stat stat_p;
    size_t filesize;
    
    // retrieve file size
    stat(codegroups[i].filename, &stat_p);
    filesize = stat_p.st_size; 
    
    if ((filesize > 0) && !(filesize % FLASH_BLOCK))
    {
      printf("Verifying checksum for '%s' (%.8X to %.8X)\n", codegroups[i].filename, codegroups[i].address, codegroups[i].address + filesize -1);
      
      if (!flash_cmd_RQRCS(dev_handle, codegroups[i].address, filesize))
        return RET_ERR;
    }
  }
    
  return RET_OK;
}


/** Flash a file to handset
 */
int flash_file(p2k_dev_handle *dev_handle, const char* filename, const size_t flashbase, const size_t skip)
{    
  struct stat stat_p;
  size_t filesize;
  size_t flashsize;
  FILE *fp;
  size_t donebytes = 0;
  size_t doneblk = 0;
  struct timeval tstart;
  double duration;
  double speed;

  // get file properties
  stat(filename, &stat_p);          

  // calculate sizes
  filesize = stat_p.st_size; 
  if (skip > filesize) {
    printf("Cannot skip %i bytes of %i bytes file!\n", skip, filesize);
    return RET_ERR;
  }
  flashsize = filesize - skip;
  
  // open the file for reading
  fp = fopen(filename,"rb");
  if (!fp) 
  {
    printf("Could not open file '%s' for reading, bailing.\n", filename);
    return RET_ERR;  
  }
  
  printf("Flashing file '%s' to handset (%i of %i bytes)..\n", filename, flashsize, filesize);
  
  // skip beginning if necessary
  if (skip)
  {
    printf("Skipping %i bytes of file\n", skip);
    fseek(fp, skip, SEEK_CUR);
  }

  timer_time(&tstart);
  while ((donebytes < flashsize) && !feof(fp))
  {
    char data[FLASH_BLOCK];    
    size_t nread;

#ifndef DEBUG
    duration = timer_elapsed(tstart);
    speed = (duration > 0) ? (double) donebytes / duration / 1000 : 0;
    if (!(doneblk % 8))
    {
      printf("\r-> Sent %i of %i bytes (%.1f kB/s)..", donebytes, flashsize, speed);
      fflush(0);
    }
#endif
    
// ADDR COMMAND   
    if (!flash_cmd_ADDR(dev_handle, flashbase + donebytes)) 
    {
      printf("Error sending ADDR command.\n");          
      return RET_ERR;
    }
    
// BIN COMMAND        
    nread = fread(data, 1, FLASH_BLOCK, fp);
    
    // send command
    if (!flash_cmd_BIN(dev_handle, data, nread))
    {
      printf("Error sending BIN command.\n");      
      return RET_ERR;
    }
    
    donebytes += nread;
    doneblk++;
  }    
  fclose(fp);

#ifndef DEBUG    
  printf("\r");
#endif
  duration = timer_elapsed(tstart);
  speed = (duration > 0) ? (double) donebytes / duration / 1000 : 0;
  printf("-> Sent %i of %i bytes (%.1f kB/s).\n", donebytes, flashsize, speed);
    
  return donebytes;
}


/** Prepare the handset for flashing
 */
int flash_prepare(p2k_dev_handle *dev_handle)
{  
  int cnt;    
      
  for (cnt = 0; cnt < 2; cnt++)
  {
    // write #1
    if (!flash_cmd_RQHW(dev_handle))
    {
      printf("Error sending RQHW command.\n");      
      return RET_ERR;
    }
    
    // write #2
    if (!flash_cmd_RQVN(dev_handle))
    {
      printf("Error sending RQVN command.\n");
      return RET_ERR;
    }      
  }  
  
  return RET_OK;
}


/** Flash a RAM loader to the handset.
 */
int flash_ramloader(p2k_dev_handle *dev_handle, const char* filename, const size_t flashbase)
{
  size_t donebytes = flash_file(dev_handle, filename, flashbase, 0);
  
// FLASH THE BOOTLOADER
  if (!donebytes)
  {
    printf("Error flashing file '%s', aborting RAMLOADER loading!", filename);
    return RET_ERR;
  }

  
// RQRC COMMAND
    
  // now we need to commit to memory
  if (!flash_cmd_RQRC(dev_handle, flashbase, donebytes))
  {
    printf("Error sending RQRC command.\n");  
    return RET_ERR;
  }    
  
// JUMP COMMAND
  if (!flash_cmd_JUMP(dev_handle, flashbase + 0x10))
  {
    printf("Error sending JUMP command.\n");
    return RET_ERR;
  }

  
// REBOOT
  printf("Sleeping for reboot..\n");    
  fflush(0);
  sleep(1);
  
  if ( !flash_cmd_RQVN(dev_handle) ||
       !flash_cmd_RQSN(dev_handle) ||
       !flash_cmd_RQSF(dev_handle) )
  {
    printf("Error during post-JUMP commands.\n");
    return RET_ERR;
  }
  
  if (!flash_cmd_UPGRADE(dev_handle))
  {
    printf("Error sending UPGRADE command.\n");
    return RET_ERR;
  }
    
  if (!flash_cmd_ERASE(dev_handle))
  {
    printf("Error sending ERASE command.\n");
    return RET_ERR;
  }
    
  return donebytes;
}


/** Read an entry from a flash catalog.
 */
int flashcat_read_entry(FILE *fp, flashcat_entry *entry)
{
  return (fscanf(fp, "%i\t%s\n", &entry->address, entry->filename) == 2) ? RET_OK : RET_ERR;
}


/** Write an entry to a flash catalog.
 */
int flashcat_write_entry(FILE *fp, flashcat_entry *entry)
{
  fprintf(fp, "%i\t%s\n", entry->address, entry->filename);  
  return RET_OK;
}


/** Split a 1FF file into a catalog and several binary files.
 *
 * @param shxfile
 */
int flash_split_1ff(const char *shxfile)
{
  char rootname[MAX_PATH];
  
  struct shx_header hdr;
  FILE *fp;
  FILE *fpc;
  unsigned char S_type;  
  struct shx_srecord S_record;
  int S_num = 0;  
  // TODO : support addresses other than 32bit long
  const unsigned char S_data = '3';
  flashcat_entry entry;
  
  root_name(shxfile, rootname);
  
  fp = fopen(shxfile, "rb");
  if (!fp)
  {
    printf("Unable to open file '%s'", shxfile);
    return RET_ERR;
  }


  // Read SHX header
  fread(&hdr, sizeof(hdr), 1, fp);    
  print_shx_info(&hdr);

       
  // Find first S0
  S_type = srecord_find(fp, 0);
  if (!S_type) {
    printf("Did not find initial S-Record, bailing!");
    return RET_ERR;
  }
  
  // write out catalog
  sprintf(entry.filename, "%s_catalog", rootname);
  fpc = fopen(entry.filename, "w");
  if (!fpc)
  {
    printf("Unable to open file '%s' for writing", entry.filename);
    return RET_ERR;
  }

  // process the binary blocks, separated by S0  
  while (S_type == '0')
  { 
    int S_started = 0;
    int cnt = 0;
    FILE *fpo;        

    if (S_num == 0) {
      sprintf(entry.filename, "%s_RAMDLD.smg", rootname);
    } else {     
      sprintf(entry.filename, "%s_CG%i.smg", rootname, S_num);
    }
    
    printf("\nStarting new codegroup\n");
    S_record = srecord_read(fp, S_type);     

 // READ S3 records containing the binary code  
 
    fpo = fopen(entry.filename, "wb");
    if (!fpo)
    {
      printf("Unable to open file '%s' for writing", entry.filename);
      return RET_ERR;
    }
 
    S_type = srecord_find(fp, 0);
    if (S_type != S_data)
    {
      printf("Current only S3 (32bit address) data records are supported!");
      return RET_ERR;
    }
    
    while (S_type == S_data)
    {
      S_record = srecord_read(fp, S_type);   
      
      if (!S_started)
      {
        entry.address = S_record.address;        
        flashcat_write_entry(fpc, &entry);
        printf("-> binary file : %s\n", entry.filename);        
        printf("-> base address : %.8X\n", entry.address);
        S_started = 1;
      }
 
      fwrite(S_record.data, S_record.datalen, 1, fpo);    
      cnt += S_record.datalen;

      S_type = srecord_find(fp, 0);
    }
  
    fclose(fpo);  
   
// SKIP UNTIL NEXT S0 
    while (S_type && (S_type != '0'))
    {
      S_record = srecord_read(fp, S_type);     
      S_type = srecord_find(fp, 0);
    }
    
    S_num ++;
    
  }
  
  fclose(fpc);      
  fclose(fp);
  
  return RET_OK;
}


/** Flash a flash catalog (generated from a 1FF file by flash-split).
 */
int flashcat_flash(p2k_dev_handle *flash_handle, const char *catalogfile, const size_t cg0break)
{
  flashcat_entry entry;
  flashcat_entry codegroups[MAX_CAT_SIZE];
  size_t cgcnt = 0;
  size_t cg0pos = 0;
  FILE *fpc;
  char cg0file[] = "flashCG0.tmp";
  
  // open catalog
  printf("Opening flash catalog '%s'\n", catalogfile);
  fpc = fopen(catalogfile, "r");
  if (!fpc)
  {
    printf("Unable to open catalog '%s', aborting.\n", catalogfile);
    return 1;
  }      
  
  // prepare for flash    
  printf("Preparing for flash\n");  
  if (!flash_prepare(flash_handle))
  {
    printf("Error while preparing for flash, aborting.\n");
    fclose(fpc);
    return RET_ERR;
  }

  // flash RAM loader
  if (!flashcat_read_entry(fpc, &entry))
  {
    printf("Malformatted catalog file, aborting.\n");
    fclose(fpc);    
    return RET_ERR;
  }
  
  printf("Flashing RAM loader '%s' to %.8X\n", entry.filename, entry.address);  
  if (!flash_ramloader(flash_handle, entry.filename, entry.address))
  {
    printf("Error while flashing RAM loader, aborting.\n");
    fclose(fpc);    
    return RET_ERR;
  } else {
    printf("RAM loader loaded OK\n");
  }
    
  
  // flash codegroups  
  while (flashcat_read_entry(fpc, &entry))
  {
    size_t cgskip = 0;
    size_t flashed;

    // check we have not exceeded maximum catalog size
    if (cgcnt >= MAX_CAT_SIZE)
    {
      printf("Exceeded the maximum catalog size of %i entries\n", MAX_CAT_SIZE);
      fclose(fpc);      
      return RET_ERR;
    }
    
    codegroups[cgcnt] = entry;
    
    // the first codegroup is handled differently
    if ((cg0break > 0) && (cgcnt == 0))
    {
      FILE *fpin;
      FILE *fpout;
      char cg0buff[4096];
      size_t nread;

      cg0pos = entry.address;

      fpin = fopen(entry.filename, "rb");
      if (!fpin)
      {
        printf("Could not open CG0 file '%s' for reading.\n", entry.filename); 
        fclose(fpc);              
        return RET_ERR;
      }
      
      nread = fread(cg0buff, 1, cg0break, fpin); 
      fclose(fpin);
      
      fpout = fopen(cg0file, "wb");
      if (!fpout)
      {
        printf("Could not open file '%s' for writing.\n", cg0file); 
        fclose(fpc);              
        return RET_ERR;
      }
      
      fwrite(cg0buff, cg0break, 1, fpout); 
      fclose(fpout);
      
      cgskip = cg0break;
    }
    
    printf("Flashing codegroup '%s' to %.8X\n", entry.filename, entry.address + cgskip);        
    flashed = flash_file(flash_handle, entry.filename, entry.address + cgskip, cgskip);
      
    if (!flashed)
    {
      printf("Error while flashing code group, aborting.\n");
      return RET_ERR;
    }
    
    cgcnt++;    
  }
  fclose(fpc);        
  
// CALCULATE CHECKSUMS

  printf("Checking checksums\n");
  if (!flash_checksum(flash_handle, codegroups, cgcnt))
  {
    printf("Checksum calculation failed.\n");
    return RET_ERR;
  }

  
// FINALISE CG0  
  if ((cg0break > 0) && (cg0pos > 0))
  {
    printf("Finalising CG0 (%.8X)\n", cg0pos);
    if (!flash_file(flash_handle, cg0file, cg0pos, 0))
    {
      printf("Failed to finalise CG0.\n");
      return RET_ERR;
    }
    remove(cg0file);
  }
  
  
// REBOOT PHONE
  printf("Restarting phone\n");
  flash_cmd_RESTART(flash_handle);
  
  printf("Done flashing.\n");
  return RET_OK;
}


/** Printout information about a SHX file.
 */
void print_shx_info(const struct shx_header *hdr)
{
  printf("Image type : %s\n", hdr->type);
  printf("Image v1   : %s\n", hdr->v1);
  printf("Image v2   : %s\n", hdr->v2);
  printf("Image date : %s\n", hdr->date);
  printf("Image info : %s\n", hdr->info);
  printf("Image v3   : %s\n", hdr->v3);
}
