/*
 * flash-tools - a set of tools for flashing handsets
 * Copyright (C) 2005 Jeremy Lain <jeremy.laine@m4x.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
#include "p2klib.h"
#include "p2kdev.h"
#include <stdio.h>
#include <string.h>


/** Find all currently connected P2k devices.
 */
size_t p2kdev_find(short vendor, short product, struct usb_device *devs[], size_t maxdevs)
{
  struct usb_bus *busses, *bus;
  struct usb_device *dev;
  size_t found_devs = 0;
  
  busses = usb_get_busses();
  
  // go over the USB busses
  for (bus = busses; bus; bus = bus->next)  
  {
    // go over the USB devices attached to the current bus
    for (dev = bus->devices; dev; dev = dev->next)
    {
      usb_dev_handle *udev;
      
      // check if this is a known P2k device
      if ((dev->descriptor.idVendor == vendor) && (dev->descriptor.idProduct == product))
      {
        // check we can open the device
        if ((udev = usb_open(dev)) != NULL)
        {
          usb_close(udev);          
          devs[found_devs] = dev;
          found_devs++;
        }
      }
      if (found_devs == maxdevs)
      {
        return found_devs;
      }
    } 
  }
 
  return found_devs;
}


/** Open a P2k device.
 */
int p2kdev_open(p2k_dev_handle *p2k_handle, struct usb_device *dev)
{ 
  char strManufacturer[P2K_STRING_MAX], strProduct[P2K_STRING_MAX];
  
  fprintf(stderr, "Opening Bus %s Device %s.. ", dev->bus->dirname, dev->filename);
  
  // open the USB device  
  p2k_handle->usb_handle = usb_open(dev);  
  if (!p2k_handle->usb_handle)
  {
    fprintf(stderr, "ERROR\nCould not open USB device : %s\n", usb_strerror());  
    return RET_ERR;
  }

  // set configuration  
  if (usb_set_configuration(p2k_handle->usb_handle, p2k_handle->bConfigurationValue)) 
  {
    fprintf(stderr, "ERROR\nCould not set USB configuration : %s\n", usb_strerror());
    usb_close(p2k_handle->usb_handle);    
    return RET_ERR;
  } 
    
  // claim interface 
  if (usb_claim_interface(p2k_handle->usb_handle, p2k_handle->bInterfaceNumber))
  {
    fprintf(stderr, "ERROR\nCould not claim USB interface : %s\n", usb_strerror());
    usb_close(p2k_handle->usb_handle);    
    return RET_ERR; 
  }

  // clear halts
  if (p2k_handle->bEndpointIn)
    usb_clear_halt(p2k_handle->usb_handle, p2k_handle->bEndpointIn);

  if (p2k_handle->bEndpointOut)
    usb_clear_halt(p2k_handle->usb_handle, p2k_handle->bEndpointOut);

  // print out some information about the device
  usb_get_string_simple(p2k_handle->usb_handle, dev->descriptor.iManufacturer, strManufacturer, P2K_STRING_MAX);
  usb_get_string_simple(p2k_handle->usb_handle, dev->descriptor.iProduct, strProduct, P2K_STRING_MAX);
  fprintf(stderr, "OK\nOpened %s %s\n", strManufacturer, strProduct);
      
  return RET_OK;
}


/** Open a P2k data logging device.
 */
int p2kdev_open_datalog(p2k_dev_handle *p2k_handle, struct usb_device *dev)
{
  p2k_handle->bConfigurationValue = 1;
  p2k_handle->bInterfaceNumber = 6;
  p2k_handle->bEndpointIn = 0x89;
  p2k_handle->bEndpointOut = 0x00;

  return p2kdev_open(p2k_handle, dev);
}


/** Open a P2k modem device.
 */
int p2kdev_open_modem(p2k_dev_handle *p2k_handle, struct usb_device *dev)
{
  p2k_handle->bConfigurationValue = 1;
  p2k_handle->bInterfaceNumber = 1;
  p2k_handle->bEndpointIn = 0x86;
  p2k_handle->bEndpointOut = 0x05;

  return p2kdev_open(p2k_handle, dev);
}


/** Open a P2k flash device.
 */
int p2kdev_open_sflash(p2k_dev_handle *p2k_handle, struct usb_device *dev)
{
  p2k_handle->bConfigurationValue = 1;
  p2k_handle->bInterfaceNumber = 0;
  p2k_handle->bEndpointIn = 0x82;
  p2k_handle->bEndpointOut = 0x01;

  return p2kdev_open(p2k_handle, dev);
}


/** Close the P2k device.
 */
int p2kdev_close(p2k_dev_handle *dev)
{
  usb_release_interface(dev->usb_handle, dev->bInterfaceNumber);  
  usb_close(dev->usb_handle);
  return RET_OK;
}


/** Write to a P2k device.
 */
int p2kdev_write(p2k_dev_handle *p2k_handle, char *bytes, const size_t size, const int timeout)
{
#ifdef DEBUG
  fprintf(stderr, "--> writing %i bytes to endpoint %i\n", size, p2k_handle->bEndpointOut);
#endif

  int bytes_out = usb_bulk_write(p2k_handle->usb_handle, p2k_handle->bEndpointOut, bytes, size, timeout);
  if (bytes_out < 0)
  {
    fprintf(stderr, "USB write failed with status %i : %s\n", bytes_out, usb_strerror());    
    return RET_ERR;
  }
  else if (((size_t)bytes_out) != size)
  {
    fprintf(stderr, "USB short write (wrote %i / %i bytes), bailing!\n", bytes_out, size);
    return RET_ERR;
  }

#ifdef DEBUG    
  hexprint(bytes, size);
  fprintf(stderr, "--> wrote %i bytes\n", size);
#endif  

  return RET_OK;
}


/** Read from a P2k device
 */
int p2kdev_read(p2k_dev_handle *p2k_handle, char *bytes, const size_t maxsize, const int timeout)
{
#ifdef DEBUG    
  fprintf(stderr, "<-- reading from endpoint %i\n", p2k_handle->bEndpointIn);
#endif  
  int bytes_in = usb_bulk_read(p2k_handle->usb_handle, p2k_handle->bEndpointIn, bytes, maxsize, timeout);
  
  if (bytes_in >= 0)
  {
#ifdef DEBUG  
    hexprint(bytes, bytes_in);
    fprintf(stderr, "<-- read %i bytes\n", bytes_in);  
#endif    
    return bytes_in;
  } else {
    fprintf(stderr, "USB read failed with status %i : %s\n", bytes_in, usb_strerror());    
    return RET_ERR;
  }  
}


/** Switch a P2k modem to composite device.
 */
int p2kdev_modem2composite(struct usb_device *dev)
{
  p2k_dev_handle p2k_handle;

  // open flash device
  if (!p2kdev_open_modem(&p2k_handle, dev))
  {
    fprintf(stderr, "Could not open USB device, aborting.\n");
    return RET_ERR;
  }

  // build AT command
  int mode = 8;
  char cmd[32];
  sprintf(cmd, "AT+MODE=%i\r\n", mode);
  
  // send AT command
  usb_bulk_write(p2k_handle.usb_handle, p2k_handle.bEndpointOut, cmd, strlen(cmd), P2K_TIMEOUT);

  // close flash device
  p2kdev_close(&p2k_handle);

  return RET_OK;
}


