/*
 * flash-tools - a set of tools for flashing handsets
 * Copyright (C) 2005 Jeremy Lain <jeremy.laine@m4x.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
#include "srecord.h"
#include "p2klib.h"
#include <stdlib.h>

/** Print information about an S-Record
 *
 * @param s
 */
void srecord_print_info(const struct shx_srecord *s)
{
  printf("S-Record type     : S%c (count : %i, checksum : %.2x)\n", s->type, s->count, s->checksum);
  
  printf("S-Record address  : %.8X (%i bytes)\n", s->address, s->addrlen);
  
  printf("S-Record data     : %i bytes\n", s->datalen);
  hexprint(s->data, s->datalen);
}


/** Look for an S-Record in a file and return the S-Record type.
 *
 * @param fp
 * @param S_find the S-Record type to look for (0 means any)
 */
int srecord_find(FILE *fp, const unsigned char S_find)
{
  unsigned char S_record[2];  
  size_t S_pos = 0;
  int found = 0;
  
  while (!found && !feof(fp))
  {
    int nread = fread(S_record, 1, 2, fp);
   
    // check if this is an S-Record
    if ((nread == 2) && (S_record[0] == 'S') && (S_record[1] >= '0') && (S_record[1] <= '9'))
    {
      if (!S_find || (S_record[1] == S_find))
      { 
        found = 1;
      }
    }
     
    if (!found) {    
      S_pos++;
    }
  }

  if (!found) {
  //  printf("Did not find S-Record!\n");
    return RET_ERR;
  }

  //printf("Found S-Record S%c (0x%.2x) at position %i\n", S_record[1], S_record[1], S_pos);  
  return S_record[1];  
}


/** Read an S-Record from a file.
 *
 * @param fp
 * @param srectype
 */
struct shx_srecord srecord_read(FILE *fp, const unsigned char srectype)
{
  struct shx_srecord s;
  
  char tmp[1024];  
  char addrtmp[4];  
  unsigned char cksum = 0;
  size_t i;
  unsigned char eol[2];
  int nread;
  
// CHECK S-Record TYPE
  
  switch(srectype) 
  {
  case '0': case '1': case '5': case '9':
    s.addrlen = 2;
    break;
    
  case '2': case '8':
    s.addrlen = 3;
    break;
    
  case '3': case '7':
    s.addrlen = 4;
    break;
          
  default:
    printf("Unknown S-Record type %.2x, bailing!\n", srectype);
    exit(2);
    break;  
  }
  s.type = srectype;
 
  
// READ S-Record COUNT
  
  fread(tmp, 2, 1, fp);
  shx_conv_chars(tmp, (char*)&s.count, 1);
  cksum += s.count;
  
  
// READ ADDRESS (2, 3 or 4 bytes)  

  s.address = 0;
  fread(tmp, 2 * s.addrlen, 1, fp);
  shx_conv_chars(tmp, addrtmp, s.addrlen);
  for (i = 0; i < s.addrlen; i++)
  {
    cksum += (unsigned char)addrtmp[i];
    s.address = (s.address << 8) + (unsigned char)addrtmp[i];
  }
  

// READ DATA
     
  // data length in bytes : count - address (addrlen bytes) - checksum (1 byte)
  s.datalen = s.count - s.addrlen - 1;
  //printf("S-Record data length : %i bytes\n", s.datalen);  
  
  // read data
  fread(tmp, s.datalen * 2, 1, fp);
  shx_conv_chars(tmp, s.data, s.datalen);  
  for (i = 0; i < s.datalen; i++)
  {
    cksum += (unsigned char)s.data[i];
  }

  
// READ AND VERIFY CHECKSUM
        
  // read checksum
  fread(tmp, 2, 1, fp);
  shx_conv_chars(tmp, (char*)&s.checksum, 1);
  
  // perform 1's complement on our calculated checksum
  cksum = 0xff - cksum;
  //printf("cksum : %.2x\n", cksum);  
  if (cksum != s.checksum) 
  {
    printf("S-Record checksum mismatch : expecting %.2x, got %.2x!", s.checksum, cksum);
  }


// READ END OF LINE  
 
  eol[0] = eol[1] = 0; 
 
  nread = fread(eol, 2, 1, fp);
  if (nread > 0)
  {
    if ((eol[0] != 0x0d) || (eol[1] != 0x0a))
    {
      printf("S-Record does not end with CRLF or EOF, bailing!\n");
      exit(2);
    }
  }

  return s;
}
